;********************************;
;     FILE PRINT UTILITY         ;
;********************************;

;********************************;
;      MISCELLANEOUS EQUATES     ;
;********************************;


MAXLN	EQU	56	;LINES/PAGE
NULLS	EQU	12	;NULLS AFTER A FORM FEED

BOOT	EQU	0000H	;REBOOT ENTRY POINT
CPM	EQU	0005H	;CPM ENTRY POINT

TBUF	EQU	0080H	;TRANSIENT PROGRAM BUFFER 
TFCB	EQU	005CH	;TRANSIENT PROGRAM FCB

OPEN	EQU	15	;OPEN FUNCTION CODE
READ	EQU	20	;READ FUNCTION CODE

;********************************;
;          MAIN LOOP             ;
;********************************;

	ORG	100H
PRINT:	LXI	SP,STACK
	LXI	D,TFCB	;POINT TO FCB
	CALL	FOPEN	;OPEN FILE
	JC	ERR	;IF ERROR, EXIT
	CALL	TOF	;PRINT HEADING

LOOP:	CALL	GETBT	;GET A BYTE
	JC	ERR	;ERROR
	CPI	1AH	;EOF?
	JZ	DONE	;YES
	CPI	0DH	;CR?
	JZ	CRET	;YES
	CPI	09H	;TAB?
	JNZ	LOOPX	;NO, CONTINUE
	LXI	H,COL	;POINT TO COLUMN
TBLP:	MVI	A,' '	;PRINT ONE SPACE
	CALL	PBYT
	MOV	A,M	;GET COLUMN
	ANI	07H	;MODULO 8
	JNZ	TBLP	;IF NOT AT TAB STOP, KEEP TYPEIN
	JMP	LOOP
LOOPX:	CALL	PBYT	;PRINT BYTE
	JMP	LOOP

CRET:	LDA	LINE	;GET LINE COUNT
	CPI	MAXLN	;PAGE OVERFLOW?
	JZ	CRET2	;YES
	INR	A	;BUMP LINE COUNT
	STA	LINE
	XRA	A
	STA	COL
	MVI	A,0DH
	JMP	LOOPX	;CONTINUE

CRET2:	MVI	A,0DH
	CALL	PBYT	;PRINT CR
	CALL	TOF
	JMP	LOOP

TOF:	MVI	A,0CH	;FORM FEED
	CALL	PBYT
	MVI	B,NULLS	;NUMBER OF RUBOUTS
TOF2:	MVI	A,7FH	;RUBOUT
	CALL	PBYT
	DCR	B	;DCREMENT COUNT
	JNZ	TOF2	;PRINT 'N' RUBOUTS
	LXI	H,FMSG	;POINT TO MESSAGE
	CALL	PSTRNG	;PRINT STRING
	LXI	H,TFCB+1	;POINT TO NAME
	MVI	B,8	;SIZE OF NAME
	CALL	PCNT	;PRINT COUNT
	MVI	A,' '	;PRINT A SPACE
	CALL	PBYT
	LXI	H,TFCB+9	;POINT TO TYPE
	MVI	B,03	;SIZE OF TYPE
	CALL	PCNT	;PRINT COUNT
	LXI	H,PMSG	;POINT TO MESSAGE
	CALL	PSTRNG	;PRINT STRING
	LDA	PAGE	;GET PAGE NUMBER
	INR	A	;BUMP IT
	STA	PAGE	;SAVE IT
	CALL	DEC	;CONVERT TO DECIMAL
	LXI	H,DECWRK	;POINT TO DEC STRING
	MVI	B,3
	CALL	PCNT	;PRINT PAGE NUMBER
	MVI	A,0DH	;PRINT CR
	CALL	PBYT
	MVI	A,0AH
	CALL	PBYT	;PRINT LF
	MVI	A,0AH
	CALL	PBYT
	XRA	A
	STA	LINE	;RESET LINE COUNT
	STA	COL	;RESET COLUMN
	RET

PBYT:	PUSH	H
	PUSH	B
	PUSH	PSW
	MOV	E,A
	MVI	C,05
	CALL	CPM	;PRINT
	POP	PSW
	CPI	20H	;NON-PRINTING?
	JC	PBY2	;YES, DONT BUMP COL
	LXI	H,COL	;INCREMENT COLUMN
	INR	M
PBY2:	MVI	C,11	;GET CONSOLE STATUS
	CALL	CPM
	CPI	00	;BREAK?
	JNZ	BOOT	;YES, DONE
	POP	B
	POP	H
	RET
PSTRNG:	MOV	A,M	;GET BYTE
	CPI	'$'	;STRING END?
	RZ		;YES, DONE
	CALL	PBYT	;PRINT BYTE
	INX	H	;BUMP POINTER
	JMP	PSTRNG	;LOOP

PCNT:	MOV	A,M	;GET BYTE
	CALL	PBYT	;PRINT IT
	INX	H	;BUMP POINTER
	DCR	B	;DECREMENT COUNT
	JNZ	PCNT
	RET

DEC:	LXI	H,DECWRK
	MVI	C,100
	CALL	DIGIT
	MVI	C,10
	CALL	DIGIT
	MVI	C,1
	CALL	DIGIT
	RET

DIGIT:	MVI	M,'0'
DI0:	SUB	C
	JM	DI1
	INR	M
	JMP	DI0
DI1:	ADD	C
	INX	H
	RET


DONE:	MVI	A,0CH	;FORM FEED
	CALL	PBYT
	MVI	B,NULLS	;RUBOUT COUNT
DLP2:	MVI	A,7FH
	CALL	PBYT
	DCR	B
	JNZ	DLP2
	JMP	BOOT	;EXIT

ERR:	LXI	D,ERMSG
	MVI	C,09H	;WRITE MSG
	CALL	CPM
	JMP	BOOT

 
;********************************;
;           F O P E N            ;
;  ROUTINE TO OPEN A DISK FILE   ;
;                                ;
;  INPUT:     DE=A(FCB)          ;
; OUTPUT:     CARRY=ERROR        ;
;********************************;

FOPEN:	MVI	C,OPEN	;OPEN CODE
	CALL	CPM	;ISSUE OPEN
	CPI	0FFH	;ERROR?
	JZ	FOERR	;YES
	XRA	A	;CLEAR CARRY
	RET
FOERR:	STC
	RET


;********************************;
;            G E T B T           ;
;   ROUTINE TO READ A BYTE       ;
;                                ;
;  OUTPUTS:     A=BYTE           ;
;               CARRY=ERROR      ;
;********************************;

GETBT:	LXI	H,TBUF+128
	XCHG		;BUFFER END ADDR. IN DE
	LHLD	INPTR	;CURRENT POINTER IN HL
	CALL	CPHL	;TEST FOR END OF BUFFER
	JZ	GETB2	;YES, READ
GETB1:	MOV	A,M	;GET BYTE
	INX	H	;BUMP POINTER
	SHLD	INPTR	;SAVE POINTER
	ORA	A	;RESET CARRY
	RET
GETB2:	MVI	C,READ	;READ CODE
	LXI	D,TFCB	;FCB ADDRESS
	CALL	CPM	;ISSUE READ
	CPI	00	;ERROR?
	JNZ	IERR	;YES
	LXI	H,TBUF	;RESET BUFFER POINTER
	SHLD	INPTR
	JMP	GETB1	;CONTINUE
IERR:	STC
	RET

;********************************;
;   MISCELLANEOUS SUBROUTINES    ;
;********************************;

;********************************;
;             C P H L            ;
;  ROUTINE TO COMPARE HL VS DE   ;
;********************************;

CPHL:	MOV	A,H
	CMP	D
	RNZ
	MOV	A,L
	CMP	E
	RET

;********************************;
;             D A T A            ;
;********************************;

COL:	DB	0	;COLUMN COUNTER
LINE:	DB	0	;LINE COUNTER
PAGE:	DB	0	;PAGE COUNTER

FMSG:	DB	0DH,0AH,'FILE: $'
PMSG:	DB	'     PAGE $'
ERMSG:	DB	'ERROR',0DH,0AH,'$'
	DS	32
STACK	EQU	$
INPTR:	DW	TBUF+128;INPUT POINTER
DECWRK:	DB	'000'
	END	PRINT
